/***************************************************************************
                          qt3_gui.cpp  -  description
                             -------------------
    begin                : Mon Oct 20 2003
    copyright            : (C) 2003 by Guilherme Fortunato
    email                : guilherme@linksat.com.br
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "qt3_gui.h"

#include <qdialog.h>
#include <qfiledialog.h>
#include <qtooltip.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <qbuttongroup.h>
#include <qgroupbox.h>
#include <qhbox.h>
#include <qvbox.h>
#include <qradiobutton.h>
#include <qmessagebox.h>
#include <qframe.h>
#include <qslider.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qvalidator.h>
#include <qcheckbox.h>
#include <qlayout.h>
#include <qapplication.h>
#include <qpainter.h>
#include <qcursor.h>
#include <qtimer.h>

static pthread_t config_thread;
static QApplication *qapp = NULL;

static void *config_thread_func( void *emp )
{
	qapp->exec();
	pthread_exit( NULL );
	return NULL;
}

class QT3_Interface
	: public Plugin_Graphical_Interface
{

friend class Mupen64_AudioConfig;

public:
	QT3_Interface( Plugin_Audio_Driver * );
	virtual ~QT3_Interface();
	virtual void Destroy();
	virtual void Show( DIALOG_TYPE );
	virtual void setAudioParameters( bool = false );
	virtual void getAudioParameters();
	void startQApp();

private:
	Mupen64_AudioConfig *confWindow;
	bool configured;
};

static QIntValidator b_time( 100000, 3000000, NULL );
static QIntValidator b_size( 1024, 32768, NULL );
static QIntValidator p_size( 64, 2048, NULL );
static QGroupBox *group_box;
static QVBox *vbox;
static QHBox *hbox;
static QFrame *frame;
static QSlider *sliderLeft, *sliderRight;
static QLineEdit *inputBufferTime, *inputBufferSize, *inputPeriodSize;
static QButtonGroup *btnGroup;
static QLabel *label;
static QComboBox *cboChannels;
static QRadioButton *alsa_opt, *oss_opt, *disk_opt;
static QCheckBox *chkSyncFix, *chkQuery, *chkLoad;
static QPushButton *btnDefault, *btnTest, *btnLoad, *btnSave, *btnAbout, *btnOK, *btnCancel, *btnChooseFile;
static QLineEdit *inputDiskFile;
static QTimer *timer;

void CreateGUIObject( Plugin_Graphical_Interface **GUI, Plugin_Audio_Driver *drv )
{
	( ( QT3_Interface * )( *GUI ) ) = new QT3_Interface( drv );
}

void DeleteGUIObject( Plugin_Graphical_Interface **GUI )
{
	delete *GUI;
	*GUI = NULL;
}

QT3_Interface::QT3_Interface( Plugin_Audio_Driver *drv )
	: Plugin_Graphical_Interface( drv ),
	configured( false )
{
}

QT3_Interface::~QT3_Interface()
{
	this->Destroy();
}

void QT3_Interface::Show( DIALOG_TYPE )
{
	if ( qapp == NULL ) startQApp();
	if ( !configured ) {
		confWindow->Configure();
		qapp->unlock( configured = true );
		getAudioParameters();
	}
	if ( !confWindow->isVisible() ) {
		backup();
		confWindow->move( ( QApplication::desktop()->width() / 2 ) - confWindow->width() / 2,
		( QApplication::desktop()->height() >> 1 ) - confWindow->height() / 2 );
		confWindow->show();
	}
}

void QT3_Interface::Destroy()
{
	confWindow->close();
	delete confWindow;
	delete qapp;
}

void QT3_Interface::startQApp()
{
	int arg_c = 0;
	char *arg_list[1];
	arg_list[0] = NULL;
	qapp = new QApplication( arg_c, arg_list, "qt-test" );
	confWindow = new Mupen64_AudioConfig( this );
	qapp->setMainWidget( confWindow );
	qapp->connect( qapp, SIGNAL( lastWindowClosed() ), qapp, SLOT( quit() ) );
	qapp->lock();
	pthread_create( &config_thread, NULL, config_thread_func, NULL );
}

void QT3_Interface::setAudioParameters( bool testing )
{
	setDriverFilename( inputDiskFile->text().data() );

	if ( disk_opt->isChecked() ) {
		if ( !testing ) {
			getAudioDriver()->setDriver( DISK_WRITER );
			return;
		}
		alsa_opt->setChecked( true );
	}

	if ( alsa_opt->isChecked() )
		getAudioDriver()->setDriver( ALSA_LIB );
	else if ( oss_opt->isChecked() )
		getAudioDriver()->setDriver( ALSA_OSS );
	getAudioDriver()->setSync( chkSyncFix->isChecked() );
	getAudioDriver()->setQuery( chkQuery->isChecked() );
	getAudioDriver()->setVolumes( sliderLeft->value(), sliderRight->value() );
	getAudioDriver()->setChannels( cboChannels->currentText().toInt() );
	getAudioDriver()->setBufferTime( inputBufferTime->text().toInt() );
	getAudioDriver()->setBufferSize( inputBufferSize->text().toInt() );
	getAudioDriver()->setPeriodSize( inputPeriodSize->text().toInt() );
	if ( !testing )
		setAutoLoad( chkLoad->isChecked() );
}

void QT3_Interface::getAudioParameters()
{
	alsa_opt->setChecked( getAudioDriver()->getDriver( ALSA_LIB ) );
	oss_opt->setChecked( getAudioDriver()->getDriver( ALSA_OSS ) );
	disk_opt->setChecked( getAudioDriver()->getDriver( DISK_WRITER ) );
	sliderLeft->setMaxValue( getAudioDriver()->getVolumeMax() );
	sliderRight->setMaxValue( getAudioDriver()->getVolumeMax() );
	sliderLeft->setValue( getAudioDriver()->getVolumeLeft() );
	sliderRight->setValue( getAudioDriver()->getVolumeRight() );
	cboChannels->setCurrentText( QString::number( getAudioDriver()->getChannels() ) );
	inputBufferTime->setText( QString::number( getAudioDriver()->getBufferTime() ) );
	inputBufferSize->setText( QString::number( getAudioDriver()->getBufferSize() ) );
	inputPeriodSize->setText( QString::number( getAudioDriver()->getPeriodSize() ) );
	inputDiskFile->setText( getDriverFilename() );
	chkLoad->setChecked( getAutoLoad() );
	chkQuery->setChecked( getAudioDriver()->getQuery() );
	chkSyncFix->setChecked( getAudioDriver()->getSync() );
}

Mupen64_AudioConfig::Mupen64_AudioConfig( Plugin_Graphical_Interface *parent )
	: QDialog( 0, "Mupen64_Audio_Plugin_Config_Dialog", false, Qt::WDestructiveClose ),
	gui( parent )
{
	hide();
}

Mupen64_AudioConfig::~Mupen64_AudioConfig()
{
}

void Mupen64_AudioConfig::Configure()
{
	QGridLayout *grid = new QGridLayout( this, 0, 0, 0, 5 );
	setCaption( CONF_DLG_CAPTION  );
	group_box = new QGroupBox( 3, Qt::Vertical, LBL_DEVICE_LIST, this );
	int var_int = 0;
	gui->getAudioDriver()->getHardware( var_int );
	if ( var_int ) {
		while ( --var_int >= 0 )
			label = new QLabel( devices[2 - var_int], group_box );
	}
	else
		label = new QLabel( LBL_NO_DEVICE, group_box );

	grid->addWidget( group_box, 0, 0 );

	Info = new infoTip( this );

	btnGroup = new QButtonGroup( 3, Qt::Vertical, this );
	btnGroup->setRadioButtonExclusive( true );
	btnGroup->setFlat( true );
	alsa_opt = new QRadioButton( LBL_ALSA, btnGroup, "a" );
	alsa_opt->setMouseTracking( true );
	alsa_opt->installEventFilter( this );
	oss_opt = new QRadioButton( LBL_OSS, btnGroup, "b" );
	oss_opt->setMouseTracking( true );
	oss_opt->installEventFilter( this );
	disk_opt = new QRadioButton( LBL_DISK, btnGroup, "c" );
	disk_opt->setMouseTracking( true );
	disk_opt->installEventFilter( this );
	QWidget *dummy;
	dummy = new QWidget( btnGroup );
	dummy = new QWidget( btnGroup );
	hbox = new QHBox( btnGroup );
	inputDiskFile = new QLineEdit( hbox );
	btnChooseFile = new QPushButton( "Choose", hbox );
	connect( btnChooseFile, SIGNAL( clicked() ), this, SLOT( showFileDialog() ) );
	grid->addWidget( btnGroup, 1, 0 );

	frame = new QFrame( this );
	frame->setFrameStyle( QFrame::Sunken | QFrame::HLine );
	frame->setLineWidth( 5 );
	grid->addWidget( frame, 2, 0 );

	hbox = new QHBox( this );
	hbox->setSpacing( 5 );
	vbox = new QVBox( hbox );
	label = new QLabel( "Left vol.", vbox, "d" );
	label->setMouseTracking( true );
	label->installEventFilter( this );
	sliderLeft = new QSlider( 0, 100, 1, 50, Qt::Vertical, vbox );
	sliderLeft->setTracking( false );
	vbox = new QVBox( hbox );
	label = new QLabel( "Right vol.", vbox, "d" );
	label->setMouseTracking( true );
	label->installEventFilter( this );
	sliderRight = new QSlider( 0, 100, 1, 50, Qt::Vertical, vbox );
	sliderRight->setTracking( false );

	vbox = new QVBox( hbox );
	label = new QLabel( LBL_CHANNELS, vbox );
	label = new QLabel( LBL_BUFFER_TIME, vbox );
	label = new QLabel( LBL_BUFFER_SIZE, vbox );
	label = new QLabel( LBL_PERIOD_SIZE, vbox );

	vbox = new QVBox( hbox );
	cboChannels = new QComboBox( false, vbox, "e" );
	var_int = 0;
	do {
		cboChannels->insertItem( cboValues[var_int], var_int );
	} while ( cboValues[++var_int] != NULL );
	cboChannels->setMouseTracking( true );
	cboChannels->installEventFilter( this );

	inputBufferTime = new QLineEdit( vbox, "f" );
	inputBufferTime->setMouseTracking( true );
	inputBufferTime->installEventFilter( this );
	inputBufferTime->setMaxLength( 7 );
	inputBufferTime->setValidator( &b_time );
	inputBufferSize = new QLineEdit( vbox, "g" );
	inputBufferSize->setMouseTracking( true );
	inputBufferSize->installEventFilter( this );
	inputBufferSize->setMaxLength( 5 );
	inputBufferSize->setValidator( &b_size );
	inputPeriodSize = new QLineEdit( vbox, "h" );
	inputPeriodSize->setMouseTracking( true );
	inputPeriodSize->installEventFilter( this );
	inputPeriodSize->setMaxLength( 4 );
	inputPeriodSize->setValidator( &p_size );
	grid->addWidget( hbox, 3, 0 );

	frame = new QFrame( this );
	frame->setFrameStyle( QFrame::Sunken | QFrame::HLine );
	frame->setLineWidth( 5 );
	grid->addWidget( frame, 4, 0 );

	vbox = new QVBox( this );
	chkSyncFix = new QCheckBox( LBL_SYNC, vbox, "i" );
	chkSyncFix->setMouseTracking( true );
	chkSyncFix->installEventFilter( this );
	chkQuery = new QCheckBox( LBL_QUERY, vbox, "j" );
	chkQuery->setMouseTracking( true );
	chkQuery->installEventFilter( this );
	chkLoad = new QCheckBox( LBL_LOAD, vbox, "k" );
	chkLoad->setMouseTracking( true );
	chkLoad->installEventFilter( this );
	grid->addWidget( vbox, 5, 0 );

	frame = new QFrame( this );
	frame->setFrameStyle( QFrame::Sunken | QFrame::HLine );
	frame->setLineWidth( 5 );
	grid->addWidget( frame, 6, 0 );

	vbox = new QVBox( this );
	hbox = new QHBox( vbox );

	btnDefault = new QPushButton( "&Default", hbox, "l" );
	btnDefault->setMouseTracking( true );
	btnDefault->installEventFilter( this );
	connect( btnDefault, SIGNAL( clicked() ), this, SLOT( loadDefaults() ) );
	btnTest = new QPushButton( "&Test", hbox, "m" );
	btnTest->setMouseTracking( true );
	btnTest->installEventFilter( this );
	connect( btnTest, SIGNAL( clicked() ), this, SLOT( begin_testing() ) );
	btnLoad = new QPushButton( "&Load", hbox, "n" );
	btnLoad->setMouseTracking( true );
	btnLoad->installEventFilter( this );
	connect( btnLoad, SIGNAL( clicked() ), this, SLOT( loadConfigurationFromFile() ) );
	btnSave = new QPushButton( "&Save", hbox, "o" );
	btnSave->setMouseTracking( true );
	btnSave->installEventFilter( this );
	connect( btnSave, SIGNAL( clicked() ), this, SLOT( saveConfigurationToFile() ) );
	btnAbout = new QPushButton( "&About", hbox );
	connect( btnAbout, SIGNAL( clicked() ), this, SLOT( showAbout() ) );

	hbox = new QHBox( vbox );
	btnOK = new QPushButton( "&OK", hbox );
	btnOK->setDefault( true );
	connect( btnOK, SIGNAL( clicked() ), this, SLOT( accept() ) );
	btnCancel = new QPushButton( "&Cancel", hbox );
	connect( btnCancel, SIGNAL( clicked() ), this, SLOT( reject() ) );
	grid->addWidget( vbox, 7, 0 );
}

bool Mupen64_AudioConfig::eventFilter( QObject *o, QEvent *e )
{
	if ( e->type() != QEvent::MouseMove ) return false;
	switch ( o->name()[0] ) {
		case 'a': Info->display( ALSA_TIP ); break;
		case 'b': Info->display( OSS_TIP ); break;
		case 'c': Info->display( DISK_TIP ); break;
		case 'd': Info->display( VOL_TIP ); break;
		case 'e': Info->display( CHANNEL_TIP ); break;
		case 'f': Info->display( TIME_TIP ); break;
		case 'g': Info->display( SIZE_TIP ); break;
		case 'h': Info->display( PERIOD_TIP ); break;
		case 'i': Info->display( SYNC_TIP ); break;
		case 'j': Info->display( QUERY_TIP ); break;
		case 'k': Info->display( LOAD_TIP ); break;
		case 'l': Info->display( DEFAULT_TIP ); break;
		case 'm': Info->display( TEST_TIP ); break;
		case 'n': Info->display( SAVE_TIP ); break;
		case 'o': Info->display( LOAD_TIP ); break;
		default: Info->hide(); return false;
	}
	return true;
}

void Mupen64_AudioConfig::showFileDialog()
{
	QFileDialog *fileDlg = new QFileDialog( getenv( "PWD" ), QString::null, this, "Disk_out_File_Dialog", true );
	fileDlg->setCaption( FILE_DLG_CAPTION );
	fileDlg->setViewMode( QFileDialog::Detail );
	fileDlg->setMode( QFileDialog::AnyFile );
	if ( fileDlg->exec() == QDialog::Accepted )
		inputDiskFile->setText( fileDlg->selectedFile() );
	delete fileDlg;
}

void Mupen64_AudioConfig::showAbout()
{
	QMessageBox *aboutBox = new QMessageBox( ABOUT_DLG_CAPTION,
		QString( "%1 %2 %3" )
		.arg( PLUGIN_NAME ).arg( VERSION ).arg( ABOUT_TEXT ),
		QMessageBox::Information,  QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton, this, "About_Box", this );
	aboutBox->exec();
	delete aboutBox;
}

void Mupen64_AudioConfig::accept()
{
	Info->hide();
	hide();
	gui->cancelConfig( true );
	gui->setAudioParameters();
}

void Mupen64_AudioConfig::reject()
{
	Info->hide();
	hide();
	gui->cancelConfig();
}

void Mupen64_AudioConfig::begin_testing()
{
	if ( !gui->isPlaying() ) {
		gui->setAudioParameters( true );
		btnTest->setText( "&Stop" );
	}
	else
		btnTest->setText( "&Test" );
	gui->playTestFile();
}

void Mupen64_AudioConfig::loadDefaults()
{
	gui->loadDefaults();
}

void Mupen64_AudioConfig::loadConfigurationFromFile()
{
	gui->LoadConfiguration();
	gui->getAudioParameters();
}

void Mupen64_AudioConfig::saveConfigurationToFile()
{
	gui->setAudioParameters();
	gui->SaveConfiguration();
}

Mupen64_AudioConfig::infoTip::infoTip( QWidget *window )
    : QFrame( 0, 0, Qt::WStyle_Customize | Qt::WDestructiveClose | Qt::WX11BypassWM
			| Qt::WStyle_StaysOnTop | Qt::WStyle_Tool | Qt::WStyle_NoBorder ),
    parentWindow( window )
{
	setBackgroundMode( Qt::NoBackground );
	setPalette( QToolTip::palette() );
	textLabel = new QLabel( this );
	textLabel->setAlignment( Qt::AlignJustify | Qt::WordBreak );
	timer = new QTimer( this );
	connect( timer, SIGNAL( timeout() ), this, SLOT( hide() ) );
}

Mupen64_AudioConfig::infoTip::~infoTip()
{
	delete textLabel;
	delete timer;
}

void Mupen64_AudioConfig::infoTip::mousePressEvent( QMouseEvent * )
{
	QFrame::hide();
}

void Mupen64_AudioConfig::infoTip::paintEvent( QPaintEvent * )
{
	textLabel->hide();
	QRect r = rect();
	r.addCoords( 0, 0, -6, -6 );
	QPainter p( this );
	p.setPen( colorGroup().foreground() );
	p.drawRect( r );
	p.setPen( colorGroup().mid() );
	p.setBrush( colorGroup().brush( QColorGroup::Background ) );
	int w = r.width();
	int h = r.height();
	p.drawRect( 1, 1, w - 2, h - 2 );
	// draw shadow
	p.setPen( colorGroup().shadow() );
	p.drawPoint( w + 5, 6 );
	p.drawLine( w + 3, 6, w + 5, 8 );
	p.drawLine( w + 1, 6, w + 5, 10 );
	int i;
	for( i = 7; i < h; i+=2 )
		p.drawLine( w, i, w + 5, i + 5 );
	for( i = w - i + h; i > 6; i-=2 )
		p.drawLine( i, h, i + 5, h + 5 );
	for( ; i > 0 ; i -= 2 )
		p.drawLine( 6, h + 6 - i, i + 5, h + 5 );
	p.setPen( colorGroup().foreground() );
	r.addCoords( 12, 8, -12, -8 );
	textLabel->show();
}

void Mupen64_AudioConfig::infoTip::display( const QString &text )
{
	static QString txt = QString::null;
	if ( isVisible() ) {
		if ( txt.compare( text ) == 0 ) return;
		hide();
	}
	txt = text;
	QRect r;
	textLabel->setText( text );
	textLabel->adjustSize();
	textLabel->move( 5, 5 );
	r.setRect( 0, 0, textLabel->width(), textLabel->height() );
	resize( r.width() + 20, r.height() + 20 );
	show();
}

void Mupen64_AudioConfig::infoTip::show()
{
	move( QCursor::pos().x(), QCursor::pos().y() - height() );
	QFrame::show();
	timer->start( 4000, true );
}

void Mupen64_AudioConfig::infoTip::hide()
{
	if ( isVisible() ) {
		timer->stop();
		QFrame::hide();
	}
}
